// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#ifndef IDYNTREE_ESTIMATION_BERDY_SPARSEMAPSOLVER_H
#define IDYNTREE_ESTIMATION_BERDY_SPARSEMAPSOLVER_H

#ifdef __DEPRECATED
  #warning <iDynTree/Estimation/BerdySparseMAPSolver.h> is deprecated. Please use <iDynTree/BerdySparseMAPSolver.h>. To disable this warning use -Wno-deprecated.
#endif

#include <iDynTree/BerdySparseMAPSolver.h>


#endif /* end of include guard: IDYNTREE_BERDY_SPARSEMAPSOLVER_H */
