// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause


#ifndef IDYNTREE_CORE_SO3UTILS_H
#define IDYNTREE_CORE_SO3UTILS_H

#ifdef __DEPRECATED
  #warning <iDynTree/Core/SO3Utils.h> is deprecated. Please use <iDynTree/SO3Utils.h>. To disable this warning use -Wno-deprecated.
#endif

#include <iDynTree/SO3Utils.h>

#endif
