// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#ifndef IDYNTREE_PRIVATE_PREPROCESSOR_UTILS_H
#define IDYNTREE_PRIVATE_PREPROCESSOR_UTILS_H

#ifdef __PRETTY_FUNCTION__
#define IDYNTREE_PRETTY_FUNCTION __PRETTY_FUNCTION__
#else 
#define IDYNTREE_PRETTY_FUNCTION __FUNCTION__
#endif

#endif /* IDYNTREE_PRIVATE_PREPROCESSOR_UTILS_H */
