#ifndef IDYNTREE_CORE_SPAN_H
#define IDYNTREE_CORE_SPAN_H

#ifdef __DEPRECATED
  #warning <iDynTree/Core/Span.h> is deprecated. Please use <iDynTree/Span.h>. To disable this warning use -Wno-deprecated.
#endif

#include <iDynTree/Span.h>

#endif
