// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#ifndef IDYNTREE_CORE_PRIVATE_UTILS_H
#define IDYNTREE_CORE_PRIVATE_UTILS_H

#ifdef __DEPRECATED
  #warning <iDynTree/Core/PrivateUtils.h> is deprecated. Please use <iDynTree/PrivateUtils.h>. To disable this warning use -Wno-deprecated.
#endif

#include <iDynTree/PrivateUtils.h>

#endif
