// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#ifndef IDYNTREE_MODEL_JACOBIANS_H
#define IDYNTREE_MODEL_JACOBIANS_H

#ifdef __DEPRECATED
  #warning <iDynTree/Model/Jacobians.h> is deprecated. Please use <iDynTree/Jacobians.h>. To disable this warning use -Wno-deprecated.
#endif

#include <iDynTree/Jacobians.h>


#endif
