// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#include <iDynTree/Indices.h>

namespace iDynTree
{

LinkIndex LINK_INVALID_INDEX  = -1;
std::string LINK_INVALID_NAME = "";

JointIndex JOINT_INVALID_INDEX = -1;
std::string JOINT_INVALID_NAME = "";

FrameIndex FRAME_INVALID_INDEX = -1;
std::string FRAME_INVALID_NAME = "";

DOFIndex   DOF_INVALID_INDEX   = -1;
std::string DOF_INVALID_NAME   = "";

TraversalIndex TRAVERSAL_INVALID_INDEX = -1;


}
