// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#ifndef IDYNTREE_ALL_SENSORS_TYPE_H
#define IDYNTREE_ALL_SENSORS_TYPE_H

#include <iDynTree/SixAxisForceTorqueSensor.h>
#include <iDynTree/AccelerometerSensor.h>
#include <iDynTree/GyroscopeSensor.h>
#include <iDynTree/ThreeAxisAngularAccelerometerSensor.h>
#include <iDynTree/ThreeAxisForceTorqueContactSensor.h>

#endif
