// SPDX-FileCopyrightText: Fondazione Istituto Italiano di Tecnologia (IIT)
// SPDX-License-Identifier: BSD-3-Clause

#include <iDynTree/AttitudeEstimator.h>


iDynTree::IAttitudeEstimator::~IAttitudeEstimator()
{
}
